@@ -11,6 +11,7 @@ Requirements:
     Place katex.min.css and katex.min.js in the same directory as this script
 """
 
+import atexit
 import hashlib
 import json
 import os
@@ -19,6 +20,8 @@ import re
 import sqlite3
 import threading
 import unittest
+import weakref
+from concurrent.futures import ThreadPoolExecutor
 from dataclasses import dataclass
 from typing import List, Optional
 
@@ -125,10 +128,6 @@ equation_cache = EquationCache()
 # --- End SQLite Cache Implementation ---
 
 
-# Thread-local storage for Playwright and browser instances
-_thread_local = threading.local()
-
-
 @dataclass
 class BoundingBox:
     x: float
@@ -158,44 +157,80 @@ def get_equation_hash(equation, bg_color="white", text_color="black", font_size=
     return hashlib.sha1(params_str.encode("utf-8")).hexdigest()
 
 
-def init_browser():
-    """
-    Initialize the Playwright and browser instance for the current thread if not already done.
-    """
-    if not hasattr(_thread_local, "playwright"):
-        _thread_local.playwright = sync_playwright().start()
-        _thread_local.browser = _thread_local.playwright.chromium.launch()
+# Thread-local storage for browser instances in the executor threads
+_thread_local = threading.local()
+
+# Global thread pool executor with a fixed number of threads
+# Each thread will maintain its own Playwright instance
+_render_executor = ThreadPoolExecutor(max_workers=8, thread_name_prefix="playwright-render")
 
 
-def get_browser():
-    """
-    Return the browser instance for the current thread.
-    """
-    init_browser()
-    return _thread_local.browser
+def _cleanup_executor():
+    """Cleanup function to shutdown the executor on exit."""
+    _render_executor.shutdown(wait=False)
 
 
-def render_equation(
-    equation,
-    bg_color="white",
-    text_color="black",
-    font_size=24,
-    use_cache=True,
-    debug_dom=False,
-):
-    """
-    Render a LaTeX equation using Playwright and KaTeX, extract the inner-most span elements
-    along with their bounding boxes, and extract the MathML output generated by KaTeX.
-    """
-    # Calculate hash for caching.
-    eq_hash = get_equation_hash(equation, bg_color, text_color, font_size)
+# Register cleanup at exit
+atexit.register(_cleanup_executor)
 
-    # Try to load from SQLite cache.
-    if use_cache:
-        cached = equation_cache.load(eq_hash)
-        if cached is not None:
-            return cached
 
+def _cleanup_playwright(playwright, browser):
+    print("Cleaning up", playwright)
+    try:
+        browser.close()
+    except Exception:
+        pass
+    try:
+        playwright.stop()
+    except Exception:
+        pass
+
+
+class _BrowserOwner:
+    def __init__(self):
+        p = sync_playwright().start()
+        b = p.chromium.launch()
+        self.p = p
+        self.browser = b
+        self._closed = False
+        # Important: don't capture `self` or globals in the finalizer
+        self._finalizer = weakref.finalize(self, _cleanup_playwright, p, b)
+
+    def close_now(self):
+        if not self._closed:
+            self._closed = True
+            self._finalizer()  # idempotent; runs at most once
+
+
+def _get_thread_local_browser():
+    """Get or create a browser instance for the current thread."""
+    owner = getattr(_thread_local, "owner", None)
+    if owner is None:
+        owner = _BrowserOwner()
+        _thread_local.owner = owner
+    return owner
+
+
+def _render_in_executor(equation, bg_color, text_color, font_size, use_cache, debug_dom, eq_hash):
+    """
+    Function to be run in the executor thread pool.
+    Each thread maintains its own Playwright instance.
+    """
+    owner = _get_thread_local_browser()
+    ctx = owner.browser.new_context(viewport={"width": 800, "height": 400})
+    try:
+        return _do_render(ctx, equation, bg_color, text_color, font_size, debug_dom)
+    finally:
+        try:
+            ctx.close()
+        except Exception:
+            pass
+
+
+def _do_render(context, equation, bg_color, text_color, font_size, debug_dom):
+    """
+    Internal rendering function that uses a provided browser context.
+    """
     # Escape the equation for use in a JavaScript string.
     escaped_equation = json.dumps(equation)
 
@@ -207,11 +242,8 @@ def render_equation(
     if not os.path.exists(katex_css_path) or not os.path.exists(katex_js_path):
         raise FileNotFoundError(f"KaTeX files not found. Please ensure katex.min.css and katex.min.js are in {script_dir}")
 
-    # Get the browser instance for the current thread.
-    browser = get_browser()
-
     # Create a new page.
-    page = browser.new_page(viewport={"width": 800, "height": 400})
+    page = context.new_page()
 
     # Basic HTML structure for rendering.
     page_html = f"""
@@ -242,7 +274,7 @@ def render_equation(
     page.set_content(page_html)
     page.add_style_tag(path=katex_css_path)
     page.add_script_tag(path=katex_js_path)
-    page.wait_for_load_state("networkidle")
+    page.wait_for_load_state("networkidle", timeout=0)
 
     katex_loaded = page.evaluate("typeof katex !== 'undefined'")
     if not katex_loaded:
@@ -275,14 +307,11 @@ def render_equation(
     if error_message:
         print(f"Error rendering equation: '{equation}'")
         print(error_message)
-        # Cache the error result so we don't retry it next time.
-        rendered_eq = RenderedEquation(mathml=error_message, spans=[], error=error_message)
-        if use_cache:
-            equation_cache.save(eq_hash, rendered_eq)
+        # Return error result
         page.close()
-        return rendered_eq
+        return RenderedEquation(mathml=error_message, spans=[], error=error_message)
 
-    page.wait_for_selector(".katex", state="attached")
+    page.wait_for_selector(".katex", state="attached", timeout=0)
 
     if debug_dom:
         katex_dom_html = page.evaluate(
@@ -327,11 +356,11 @@ def render_equation(
     # Extract MathML output (if available) from the KaTeX output.
     mathml = page.evaluate(
         """
-    () => {
-        const mathElem = document.querySelector('.katex-mathml math');
-        return mathElem ? mathElem.outerHTML : "";
-    }
-    """
+        () => {
+            const mathElem = document.querySelector('.katex-mathml math');
+            return mathElem ? mathElem.outerHTML : "";
+        }
+        """
     )
 
     page.close()
@@ -352,9 +381,43 @@ def render_equation(
         ],
     )
 
-    # Save the successfully rendered equation to the SQLite cache.
+    return rendered_eq
+
+
+def render_equation(
+    equation,
+    bg_color="white",
+    text_color="black",
+    font_size=24,
+    use_cache=True,
+    debug_dom=False,
+):
+    """
+    Render a LaTeX equation using Playwright and KaTeX, extract the inner-most span elements
+    along with their bounding boxes, and extract the MathML output generated by KaTeX.
+
+    This function uses a ThreadPoolExecutor with a fixed number of threads to prevent
+    resource leaks from unbounded thread creation.
+    """
+    # Calculate hash for caching.
+    eq_hash = get_equation_hash(equation, bg_color, text_color, font_size)
+
+    # Try to load from SQLite cache.
     if use_cache:
+        cached = equation_cache.load(eq_hash)
+        if cached is not None:
+            return cached
+
+    # Submit the rendering task to the thread pool executor
+    future = _render_executor.submit(_render_in_executor, equation, bg_color, text_color, font_size, use_cache, debug_dom, eq_hash)
+
+    # Wait for the result
+    rendered_eq = future.result()
+
+    # Save to cache if successful and caching is enabled
+    if use_cache and rendered_eq and not rendered_eq.error:
         equation_cache.save(eq_hash, rendered_eq)
+
     return rendered_eq
 
 
