@@ -11,10 +11,11 @@ import os
 import random
 import re
 import shutil
+import ssl
 import sys
 import tempfile
 import time
-from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
+from concurrent.futures import ThreadPoolExecutor, as_completed
 from concurrent.futures.process import BrokenProcessPool
 from dataclasses import dataclass
 from functools import cache
@@ -37,7 +38,7 @@ from olmocr.data.renderpdf import render_pdf_to_base64png
 from olmocr.filter.filter import Language, PdfFilter
 from olmocr.image_utils import convert_image_to_pdf_bytes, is_jpeg, is_png
 from olmocr.metrics import MetricsKeeper, WorkerTracker
-from olmocr.prompts import PageResponse, build_no_anchoring_yaml_prompt
+from olmocr.prompts import PageResponse, build_no_anchoring_v4_yaml_prompt
 from olmocr.prompts.anchor import get_anchor_text
 from olmocr.s3_utils import (
     download_directory,
@@ -49,7 +50,7 @@ from olmocr.s3_utils import (
 )
 from olmocr.train.dataloader import FrontMatterParser
 from olmocr.version import VERSION
-from olmocr.work_queue import LocalWorkQueue, S3WorkQueue, WorkQueue
+from olmocr.work_queue import LocalBackend, S3Backend, WorkQueue
 
 # Initialize logger
 logger = logging.getLogger(__name__)
@@ -71,6 +72,7 @@ console_handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(level
 logger.addHandler(file_handler)
 logger.addHandler(console_handler)
 server_logger.addHandler(file_handler)
+server_logger.addHandler(console_handler)
 
 # Quiet logs from pypdf
 logging.getLogger("pypdf").setLevel(logging.ERROR)
@@ -83,15 +85,11 @@ pdf_s3 = boto3.client("s3")
 metrics = MetricsKeeper(window=60 * 5)
 tracker = WorkerTracker()
 
-# Process pool for offloading cpu bound work, like calculating anchor texts, max 32 workers, otherwise it can spawn way too many workers on a big machine
-process_pool = ProcessPoolExecutor(max_workers=min(multiprocessing.cpu_count() // 2 + 1, 32), mp_context=multiprocessing.get_context("spawn"))
+pdf_render_max_workers = asyncio.BoundedSemaphore(int(float(os.environ.get("BEAKER_ASSIGNED_CPU_COUNT", max(1, multiprocessing.cpu_count() - 2)))))
 
 # Filter object, cached so it will only get loaded when/if you need it
 get_pdf_filter = cache(lambda: PdfFilter(languages_to_keep={Language.ENGLISH, None}, apply_download_spam_check=True, apply_form_check=True))
 
-# Specify a default port, but it can be overridden by args
-BASE_SERVER_PORT = 30024
-
 
 @dataclass(frozen=True)
 class PageResult:
@@ -104,12 +102,13 @@ class PageResult:
     is_fallback: bool
 
 
-async def build_page_query(local_pdf_path: str, page: int, target_longest_image_dim: int, image_rotation: int = 0) -> dict:
-    MAX_TOKENS = 4500
+async def build_page_query(local_pdf_path: str, page: int, target_longest_image_dim: int, image_rotation: int = 0, model_name: str = "olmocr") -> dict:
+    MAX_TOKENS = 8000
     assert image_rotation in [0, 90, 180, 270], "Invalid image rotation provided in build_page_query"
 
-    # Allow the page rendering to process in the background while we get the anchor text (which blocks the main thread)
-    image_base64 = await asyncio.to_thread(render_pdf_to_base64png, local_pdf_path, page, target_longest_image_dim=target_longest_image_dim)
+    # Allow the page rendering to process in the background, but limit the number of workers otherwise you can overload the system
+    async with pdf_render_max_workers:
+        image_base64 = await asyncio.to_thread(render_pdf_to_base64png, local_pdf_path, page, target_longest_image_dim=target_longest_image_dim)
 
     if image_rotation != 0:
         image_bytes = base64.b64decode(image_base64)
@@ -131,13 +130,13 @@ async def build_page_query(local_pdf_path: str, page: int, target_longest_image_
         image_base64 = base64.b64encode(buffered.getvalue()).decode("utf-8")
 
     return {
-        "model": "olmocr",
+        "model": model_name,
         "messages": [
             {
                 "role": "user",
                 "content": [
+                    {"type": "text", "text": build_no_anchoring_v4_yaml_prompt()},
                     {"type": "image_url", "image_url": {"url": f"data:image/png;base64,{image_base64}"}},
-                    {"type": "text", "text": build_no_anchoring_yaml_prompt()},
                 ],
             }
         ],
@@ -150,29 +149,44 @@ async def build_page_query(local_pdf_path: str, page: int, target_longest_image_
 # It feels strange perhaps, but httpx and aiohttp are very complex beasts
 # Ex. the sessionpool in httpcore has 4 different locks in it, and I've noticed
 # that at the scale of 100M+ requests, that they deadlock in different strange ways
-async def apost(url, json_data):
+async def apost(url, json_data, api_key=None):
     parsed_url = urlparse(url)
     host = parsed_url.hostname
-    port = parsed_url.port or 80
+    # Default to 443 for HTTPS, 80 for HTTP
+    if parsed_url.scheme == "https":
+        port = parsed_url.port or 443
+        use_ssl = True
+    else:
+        port = parsed_url.port or 80
+        use_ssl = False
     path = parsed_url.path or "/"
 
     writer = None
     try:
-        reader, writer = await asyncio.open_connection(host, port)
+        if use_ssl:
+            ssl_context = ssl.create_default_context()
+            reader, writer = await asyncio.open_connection(host, port, ssl=ssl_context)
+        else:
+            reader, writer = await asyncio.open_connection(host, port)
 
         json_payload = json.dumps(json_data)
-        request = (
-            f"POST {path} HTTP/1.1\r\n"
-            f"Host: {host}\r\n"
-            f"Content-Type: application/json\r\n"
-            f"Content-Length: {len(json_payload)}\r\n"
-            f"Connection: close\r\n\r\n"
-            f"{json_payload}"
-        )
+
+        headers = [
+            f"POST {path} HTTP/1.1",
+            f"Host: {host}",
+            f"Content-Type: application/json",
+            f"Content-Length: {len(json_payload)}",
+        ]
+
+        if api_key:
+            headers.append(f"Authorization: Bearer {api_key}")
+
+        headers.append("Connection: close")
+
+        request = "\r\n".join(headers) + "\r\n\r\n" + json_payload
         writer.write(request.encode())
         await writer.drain()
 
-        # Read status line
         status_line = await reader.readline()
         if not status_line:
             raise ConnectionError("No response from server")
@@ -194,8 +208,29 @@ async def apost(url, json_data):
         if "content-length" in headers:
             body_length = int(headers["content-length"])
             response_body = await reader.readexactly(body_length)
+        elif headers.get("transfer-encoding", "") == "chunked":
+            chunks = []
+            while True:
+                # Read chunk size line
+                size_line = await reader.readline()
+                chunk_size = int(size_line.strip(), 16)  # Hex format
+
+                if chunk_size == 0:
+                    await reader.readline()  # Read final CRLF
+                    break
+
+                chunk_data = await reader.readexactly(chunk_size)
+                chunks.append(chunk_data)
+
+                # Read trailing CRLF after chunk data
+                await reader.readline()
+
+            response_body = b"".join(chunks)
+        elif headers.get("connection", "") == "close":
+            # Read until connection closes
+            response_body = await reader.read()
         else:
-            raise ConnectionError("Anything other than fixed content length responses are not implemented yet")
+            raise ConnectionError("Cannot determine response body length")
 
         return status_code, response_body
     except Exception as e:
@@ -212,22 +247,24 @@ async def apost(url, json_data):
 
 
 async def process_page(args, worker_id: int, pdf_orig_path: str, pdf_local_path: str, page_num: int) -> PageResult:
-    COMPLETION_URL = f"http://localhost:{BASE_SERVER_PORT}/v1/chat/completions"
+    COMPLETION_URL = f"{args.server.rstrip('/')}/chat/completions"
     MAX_RETRIES = args.max_page_retries
     MODEL_MAX_CONTEXT = 16384
     TEMPERATURE_BY_ATTEMPT = [0.1, 0.1, 0.2, 0.3, 0.5, 0.8, 0.9, 1.0]
     exponential_backoffs = 0
-    local_image_rotation = 0
+    cumulative_rotation = 0  # Track cumulative rotation instead of local
     attempt = 0
     await tracker.track_work(worker_id, f"{pdf_orig_path}-{page_num}", "started")
 
     while attempt < MAX_RETRIES:
         lookup_attempt = min(attempt, len(TEMPERATURE_BY_ATTEMPT) - 1)
+
         query = await build_page_query(
             pdf_local_path,
             page_num,
             args.target_longest_image_dim,
-            image_rotation=local_image_rotation,
+            image_rotation=cumulative_rotation,
+            model_name=args.model,
         )
         # Change temperature as number of attempts increases to overcome repetition issues at expense of quality
         query["temperature"] = TEMPERATURE_BY_ATTEMPT[lookup_attempt]
@@ -238,13 +275,20 @@ async def process_page(args, worker_id: int, pdf_orig_path: str, pdf_local_path:
                 r"---\nprimary_language: (?:[a-z]{2}|null)\nis_rotation_valid: (?:True|False|true|false)\nrotation_correction: (?:0|90|180|270)\nis_table: (?:True|False|true|false)\nis_diagram: (?:True|False|true|false)\n(?:---|---\n[\s\S]+)"
             )
 
-        logger.info(f"Built page query for {pdf_orig_path}-{page_num}")
+        logger.debug(f"Built page query for {pdf_orig_path}-{page_num}")
 
         try:
-            status_code, response_body = await apost(COMPLETION_URL, json_data=query)
+            # Passing API key only for external servers that need authentication
+            if args.server and hasattr(args, "api_key"):
+                api_key = args.api_key
+            else:
+                api_key = None
+            status_code, response_body = await apost(COMPLETION_URL, json_data=query, api_key=api_key)
 
             if status_code == 400:
                 raise ValueError(f"Got BadRequestError from server: {response_body}, skipping this response")
+            elif status_code == 429:
+                raise ConnectionError(f"Too many requests, doing exponential backoff")
             elif status_code == 500:
                 raise ValueError(f"Got InternalServerError from server: {response_body}, skipping this response")
             elif status_code != 200:
@@ -273,7 +317,9 @@ async def process_page(args, worker_id: int, pdf_orig_path: str, pdf_local_path:
                 logger.info(
                     f"Got invalid_page rotation for {pdf_orig_path}-{page_num} attempt {attempt}, retrying with {page_response.rotation_correction} rotation"
                 )
-                local_image_rotation = page_response.rotation_correction
+                # Add the rotation correction to the cumulative rotation
+                cumulative_rotation = (cumulative_rotation + page_response.rotation_correction) % 360
+                logger.info(f"Cumulative rotation is now {cumulative_rotation} degrees")
                 raise ValueError(f"invalid_page rotation for {pdf_orig_path}-{page_num}")
 
             metrics.add_metrics(**{"completed_pages": 1, f"finished_on_attempt_{attempt}": 1})
@@ -358,7 +404,7 @@ async def process_pdf(args, worker_id: int, pdf_orig_path: str):
             logger.exception(f"Could not count number of pages for {pdf_orig_path}, aborting document")
             return None
 
-        logger.info(f"Got {num_pages} pages to do for {pdf_orig_path} in worker {worker_id}")
+        logger.debug(f"Got {num_pages} pages to do for {pdf_orig_path} in worker {worker_id}")
 
         if args.apply_filter and get_pdf_filter().filter_out_pdf(tf.name):
             logger.info(f"Filtering out pdf {pdf_orig_path}")
@@ -447,7 +493,14 @@ def build_dolma_document(pdf_orig_path, page_results):
         "added": datetime.datetime.now().strftime("%Y-%m-%d"),
         "created": datetime.datetime.now().strftime("%Y-%m-%d"),
         "metadata": metadata,
-        "attributes": {"pdf_page_numbers": pdf_page_spans},
+        "attributes": {
+            "pdf_page_numbers": pdf_page_spans,
+            "primary_language": [p.response.primary_language for p in page_results],
+            "is_rotation_valid": [p.response.is_rotation_valid for p in page_results],
+            "rotation_correction": [p.response.rotation_correction for p in page_results],
+            "is_table": [p.response.is_table for p in page_results],
+            "is_diagram": [p.response.is_diagram for p in page_results],
+        },
     }
     return dolma_doc
 
@@ -503,6 +556,8 @@ async def worker(args, work_queue: WorkQueue, semaphore, worker_id):
                     bucket, key = parse_s3_path(output_final_path)
                     workspace_s3.upload_file(temp_path, bucket, key)
                 else:
+                    # Ensure the results directory exists for local workspace
+                    os.makedirs(os.path.dirname(output_final_path), exist_ok=True)
                     shutil.copyfile(temp_path, output_final_path)
             finally:
                 # Clean up the temporary file
@@ -568,13 +623,13 @@ async def worker(args, work_queue: WorkQueue, semaphore, worker_id):
             semaphore.release()
 
 
-async def vllm_server_task(model_name_or_path, args, semaphore):
+async def vllm_server_task(model_name_or_path, args, semaphore, unknown_args=None):
     cmd = [
         "vllm",
         "serve",
         model_name_or_path,
         "--port",
-        str(BASE_SERVER_PORT),
+        str(args.port),
         "--disable-log-requests",
         "--uvicorn-log-level",
         "warning",
@@ -584,6 +639,8 @@ async def vllm_server_task(model_name_or_path, args, semaphore):
         str(args.tensor_parallel_size),
         "--data-parallel-size",
         str(args.data_parallel_size),
+        "--limit-mm-per-prompt",
+        '{"video": 0}',  # Disabling video encoder saves RAM that you can put towards the KV cache, thanks @charitarthchugh
     ]
 
     if args.gpu_memory_utilization is not None:
@@ -592,10 +649,15 @@ async def vllm_server_task(model_name_or_path, args, semaphore):
     if args.max_model_len is not None:
         cmd.extend(["--max-model-len", str(args.max_model_len)])
 
+    if unknown_args:
+        cmd.extend(unknown_args)
+
     proc = await asyncio.create_subprocess_exec(
         *cmd,
         stdout=asyncio.subprocess.PIPE,
         stderr=asyncio.subprocess.PIPE,
+        # OMP_NUM_THREADS needs to be 1, otherwise you could have contention if you are running multiple copies of olmOCR on a machine with several GPUS
+        env={**os.environ, "OMP_NUM_THREADS": "1"},
     )
 
     # Ensure the subprocess is terminated on exit
@@ -608,19 +670,15 @@ async def vllm_server_task(model_name_or_path, args, semaphore):
     atexit.register(_kill_proc)
 
     # Shared variables between tasks
-    last_running_req, last_queue_req = 0, 0
+    last_running_req, peak_running_req, last_queue_req = 0, 0, 0
+    running_reqs_decreased = False
     server_printed_ready_message = False
     last_semaphore_release = time.time()
 
     async def process_line(line):
-        nonlocal last_running_req, last_queue_req, last_semaphore_release, server_printed_ready_message
+        nonlocal last_running_req, last_queue_req, peak_running_req, running_reqs_decreased, last_semaphore_release, server_printed_ready_message
         server_logger.info(line)
 
-        # if the server hasn't initialized yet, log all the lines to the main logger also, so that the user
-        # can see any warnings/errors more easily
-        if not server_printed_ready_message:
-            logger.info(line)
-
         if "Detected errors during sampling" in line:
             logger.error("Cannot continue, sampling errors detected, model is probably corrupt")
             sys.exit(1)
@@ -629,12 +687,19 @@ async def vllm_server_task(model_name_or_path, args, semaphore):
             server_printed_ready_message = True
             last_semaphore_release = time.time()
 
-        match = re.search(r"Running: (\d+)", line)
-        if match:
-            last_running_req = int(match.group(1))
+        if match := re.search(r"Running: (\d+)", line):
+            current_running = int(match.group(1))
+            # Track peak running requests
+            if current_running > peak_running_req:
+                peak_running_req = current_running
+                logger.info(f"New peak running requests: {peak_running_req}")
+            # Check for negative derivative (decrease in running requests), to not overload VLLM in times of high contention
+            if current_running < last_running_req and not running_reqs_decreased:
+                running_reqs_decreased = True
+                logger.info(f"Running requests decreased: {last_running_req} -> {current_running}")
+            last_running_req = current_running
 
-        match = re.search(r"(?:Waiting|Pending):\s*(\d+)", line)
-        if match:
+        if match := re.search(r"(?:Waiting|Pending):\s*(\d+)", line):
             last_queue_req = int(match.group(1))
             logger.info(f"vllm running req: {last_running_req} queue req: {last_queue_req}")
 
@@ -650,14 +715,25 @@ async def vllm_server_task(model_name_or_path, args, semaphore):
                 logger.warning(f"Got {ex} when reading log line from inference server, skipping")
 
     async def timeout_task():
-        nonlocal last_running_req, last_queue_req, last_semaphore_release
+        nonlocal last_running_req, last_queue_req, peak_running_req, last_semaphore_release, running_reqs_decreased
         try:
             while True:
                 await asyncio.sleep(1)
-                if server_printed_ready_message and last_queue_req == 0 and time.time() - last_semaphore_release > 30 and semaphore.locked():
+
+                # Check if we should release the semaphore
+                should_release = (
+                    server_printed_ready_message
+                    and last_queue_req <= int(peak_running_req * 0.2)
+                    and time.time() - last_semaphore_release > 30
+                    and semaphore.locked()
+                    and (last_running_req == 0 or running_reqs_decreased)
+                )
+
+                if should_release:
                     semaphore.release()
+                    running_reqs_decreased = False  # Reset flag after release
                     last_semaphore_release = time.time()
-                    logger.info("Semaphore released, allowing a worker to proceed.")
+                    logger.info(f"Semaphore released at {last_running_req} running {last_queue_req} queued, peak: {peak_running_req})")
         except asyncio.CancelledError:
             pass  # Clean up if the task is cancelled
 
@@ -681,12 +757,12 @@ async def vllm_server_task(model_name_or_path, args, semaphore):
     await asyncio.gather(stdout_task, stderr_task, timeout_task, return_exceptions=True)
 
 
-async def vllm_server_host(model_name_or_path, args, semaphore):
+async def vllm_server_host(model_name_or_path, args, semaphore, unknown_args=None):
     MAX_RETRIES = 5
     retry = 0
 
     while retry < MAX_RETRIES:
-        await vllm_server_task(model_name_or_path, args, semaphore)
+        await vllm_server_task(model_name_or_path, args, semaphore, unknown_args)
         logger.warning("VLLM server task ended")
         retry += 1
 
@@ -699,15 +775,19 @@ async def vllm_server_host(model_name_or_path, args, semaphore):
         sys.exit(1)
 
 
-async def vllm_server_ready():
+async def vllm_server_ready(args):
     max_attempts = 300
     delay_sec = 1
-    url = f"http://localhost:{BASE_SERVER_PORT}/v1/models"
+    url = f"{args.server.rstrip('/')}/models"
 
     for attempt in range(1, max_attempts + 1):
         try:
+            headers = {}
+            if args.server and hasattr(args, "api_key") and args.api_key:
+                headers["Authorization"] = f"Bearer {args.api_key}"
+
             async with httpx.AsyncClient() as session:
-                response = await session.get(url)
+                response = await session.get(url, headers=headers)
 
                 if response.status_code == 200:
                     logger.info("vllm server is ready.")
@@ -838,7 +918,7 @@ def submit_beaker_job(args):
 
     # Create the experiment spec
     experiment_spec = ExperimentSpec(
-        budget="ai2/oe-data",
+        budget="ai2/oe-base",
         description=task_name,
         tasks=[
             TaskSpec(
@@ -996,7 +1076,7 @@ def print_stats(args, root_work_queue):
 
 
 async def main():
-    parser = argparse.ArgumentParser(description="Manager for running millions of PDFs through a batch inference pipeline")
+    parser = argparse.ArgumentParser(description="Manager for running millions of PDFs through a batch inference pipeline.")
     parser.add_argument(
         "workspace",
         help="The filesystem path where work will be stored, can be a local folder, or an s3 path if coordinating work with many workers, s3://bucket/prefix/ ",
@@ -1009,26 +1089,35 @@ async def main():
     )
     parser.add_argument(
         "--model",
-        help="Path where the model is located, allenai/olmOCR-7B-0725-FP8 is the default, can be local, s3, or hugging face.",
-        default="allenai/olmOCR-7B-0725-FP8",
+        help="Path where the model is located, allenai/olmOCR-2-7B-1025-FP8 is the default, can be local, s3, or hugging face.",
+        default="allenai/olmOCR-2-7B-1025-FP8",
     )
 
     # More detailed config options, usually you shouldn't have to change these
     parser.add_argument("--workspace_profile", help="S3 configuration profile for accessing the workspace", default=None)
     parser.add_argument("--pdf_profile", help="S3 configuration profile for accessing the raw pdf documents", default=None)
-    parser.add_argument("--pages_per_group", type=int, default=500, help="Aiming for this many pdf pages per work item group")
+    parser.add_argument("--pages_per_group", type=int, default=argparse.SUPPRESS, help="Aiming for this many pdf pages per work item group")
     parser.add_argument("--max_page_retries", type=int, default=8, help="Max number of times we will retry rendering a page")
     parser.add_argument("--max_page_error_rate", type=float, default=0.004, help="Rate of allowable failed pages in a document, 1/250 by default")
     parser.add_argument("--workers", type=int, default=20, help="Number of workers to run at a time")
     parser.add_argument("--apply_filter", action="store_true", help="Apply basic filtering to English pdfs which are not forms, and not likely seo spam")
     parser.add_argument("--stats", action="store_true", help="Instead of running any job, reports some statistics about the current workspace")
     parser.add_argument("--markdown", action="store_true", help="Also write natural text to markdown files preserving the folder structure of the input pdfs")
-
     parser.add_argument("--target_longest_image_dim", type=int, help="Dimension on longest side to use for rendering the pdf pages", default=1288)
     parser.add_argument("--target_anchor_text_len", type=int, help="Maximum amount of anchor text to use (characters), not used for new models", default=-1)
     parser.add_argument("--guided_decoding", action="store_true", help="Enable guided decoding for model YAML type outputs")
 
-    vllm_group = parser.add_argument_group("VLLM Forwarded arguments")
+    server_group = parser.add_argument_group("Server arguments, to specify where your VLLM inference engine is running")
+    server_group.add_argument(
+        "--server",
+        type=str,
+        help="URL of external vLLM (or other compatible provider) server (e.g., http://hostname:port/v1). If provided, skips spawning local vLLM instance",
+    )
+    server_group.add_argument("--api_key", type=str, default=None, help="API key for authenticated remote servers (e.g., DeepInfra)")
+
+    vllm_group = parser.add_argument_group(
+        "VLLM arguments", "These arguments are passed to vLLM. Any unrecognized arguments are also automatically forwarded to vLLM."
+    )
     vllm_group.add_argument(
         "--gpu-memory-utilization", type=float, help="Fraction of VRAM vLLM may pre-allocate for KV-cache " "(passed through to vllm serve)."
     )
@@ -1044,25 +1133,22 @@ async def main():
     beaker_group.add_argument(
         "--beaker_cluster",
         help="Beaker clusters you want to run on",
-        default=["ai2/jupiter-cirrascale-2", "ai2/ceres-cirrascale", "ai2/neptune-cirrascale", "ai2/saturn-cirrascale", "ai2/augusta-google-1"],
+        default=["ai2/jupiter", "ai2/ceres", "ai2/neptune", "ai2/saturn"],
     )
     beaker_group.add_argument("--beaker_gpus", type=int, default=1, help="Number of gpu replicas to run")
     beaker_group.add_argument("--beaker_priority", type=str, default="normal", help="Beaker priority level for the job")
 
-    args = parser.parse_args()
+    args, unknown_args = parser.parse_known_args()
 
     logger.info(
         "If you run out of GPU memory during start-up or get 'KV cache is larger than available memory' errors, retry with lower values, e.g. --gpu_memory_utilization 0.80  --max_model_len 16384"
     )
 
+    use_internal_server = not args.server
     global workspace_s3, pdf_s3
-    # set the global BASE_SERVER_PORT from args
-    global BASE_SERVER_PORT
-    BASE_SERVER_PORT = args.port
 
     # setup the job to work in beaker environment, load secrets, adjust logging, etc.
     if "BEAKER_JOB_NAME" in os.environ:
-        server_logger.addHandler(console_handler)
         cred_path = os.path.join(os.path.expanduser("~"), ".aws", "credentials")
         os.makedirs(os.path.dirname(cred_path), exist_ok=True)
         with open(cred_path, "w") as f:
@@ -1082,6 +1168,11 @@ async def main():
         logger.info(f"Beaker job sleeping for {sleep_time} seconds to stagger model downloads")
         await asyncio.sleep(sleep_time)
 
+    # If you specify an API key, meaning you are on a remote provider, then lower the group size default, not to overwhelm such servers
+    # and not to waste money if a group doesn't finish right away
+    if not hasattr(args, "pages_per_group"):
+        args.pages_per_group = 50 if args.api_key is not None else 500
+
     if args.workspace_profile:
         workspace_session = boto3.Session(profile_name=args.workspace_profile)
         workspace_s3 = workspace_session.client("s3")
@@ -1095,9 +1186,9 @@ async def main():
 
     # Create work queue
     if args.workspace.startswith("s3://"):
-        work_queue = S3WorkQueue(workspace_s3, args.workspace)
+        work_queue = WorkQueue(S3Backend(workspace_s3, args.workspace))
     else:
-        work_queue = LocalWorkQueue(args.workspace)
+        work_queue = WorkQueue(LocalBackend(args.workspace))
 
     if args.pdfs:
         logger.info("Got --pdfs argument, going to add to the work queue")
@@ -1175,12 +1266,20 @@ async def main():
 
     # If you get this far, then you are doing inference and need a GPU
     # check_sglang_version()
-    check_torch_gpu_available()
+    if use_internal_server:
+        check_torch_gpu_available()
 
     logger.info(f"Starting pipeline with PID {os.getpid()}")
 
     # Download the model before you do anything else
-    model_name_or_path = await download_model(args.model)
+    if use_internal_server:
+        model_name_or_path = await download_model(args.model)
+        args.server = f"http://localhost:{args.port}/v1"
+        args.model = "olmocr"  # Internal server always uses this name for the model, for supporting weird local model paths
+        logger.info(f"Using internal server at {args.server}")
+    else:
+        logger.info(f"Using external server at {args.server}")
+        model_name_or_path = None
 
     # Initialize the work queue
     qsize = await work_queue.initialize_queue()
@@ -1194,9 +1293,12 @@ async def main():
     # As soon as one worker is no longer saturating the gpu, the next one can start sending requests
     semaphore = asyncio.Semaphore(1)
 
-    vllm_server = asyncio.create_task(vllm_server_host(model_name_or_path, args, semaphore))
+    # Start local vLLM instance if not using external one
+    vllm_server = None
+    if use_internal_server:
+        vllm_server = asyncio.create_task(vllm_server_host(model_name_or_path, args, semaphore, unknown_args))
 
-    await vllm_server_ready()
+    await vllm_server_ready(args)
 
     metrics_task = asyncio.create_task(metrics_reporter(work_queue))
 
@@ -1209,14 +1311,16 @@ async def main():
     # Wait for all worker tasks to finish
     await asyncio.gather(*worker_tasks)
 
-    # Wait for server to stop
-    process_pool.shutdown(wait=False)
-
-    vllm_server.cancel()
+    # Cancel vLLM server if it was started
+    if vllm_server is not None:
+        vllm_server.cancel()
     metrics_task.cancel()
 
     # Wait for cancelled tasks to complete
-    await asyncio.gather(vllm_server, metrics_task, return_exceptions=True)
+    tasks_to_wait = [metrics_task]
+    if vllm_server is not None:
+        tasks_to_wait.append(vllm_server)
+    await asyncio.gather(*tasks_to_wait, return_exceptions=True)
 
     # Output final metrics summary
     metrics_summary = metrics.get_metrics_summary()
