@@ -31,6 +31,7 @@ Examples:
 
 import argparse
 import concurrent.futures
+import fnmatch
 import json
 import os
 import shutil
@@ -59,11 +60,20 @@ TOKENIZER_FILES = ["chat_template.json", "merges.txt", "preprocessor_config.json
 SUPPORTED_ARCHITECTURES = ["Qwen2VLForConditionalGeneration", "Qwen2_5_VLForConditionalGeneration"]
 
 # Files to exclude from copying (training-related files)
-EXCLUDED_FILES = {"optimizer.pt", "scheduler.pt", "rng_state.pth", "trainer_state.json", "training_args.bin"}
+# Supports exact matches and glob patterns
+EXCLUDED_FILES = {"optimizer.pt", "scheduler.pt", "rng_state.pth", "trainer_state.json", "training_args.bin", "*.pt", "*.pth"}
 
 s3_client = boto3.client("s3")
 
 
+def should_exclude_file(filename: str) -> bool:
+    """Check if a file should be excluded based on EXCLUDED_FILES patterns."""
+    for pattern in EXCLUDED_FILES:
+        if fnmatch.fnmatch(filename, pattern):
+            return True
+    return False
+
+
 def is_s3_path(path: str) -> bool:
     """Check if a path is an S3 path."""
     return path.startswith("s3://")
@@ -123,7 +133,7 @@ def copy_local_to_local(source_dir: str, dest_dir: str) -> None:
     files_to_copy = []
     for root, _, files in os.walk(source_dir):
         for file in files:
-            if file in EXCLUDED_FILES:
+            if should_exclude_file(file):
                 print(f"Skipping excluded file: {file}")
                 continue
             src_path = os.path.join(root, file)
@@ -164,7 +174,7 @@ def copy_s3_to_local(source_bucket: str, source_prefix: str, dest_dir: str) -> N
                 continue
 
             filename = os.path.basename(key)
-            if filename in EXCLUDED_FILES:
+            if should_exclude_file(filename):
                 print(f"Skipping excluded file: {filename}")
                 continue
 
@@ -187,7 +197,7 @@ def copy_local_to_s3(source_dir: str, dest_bucket: str, dest_prefix: str) -> Non
     upload_tasks = []
     for root, _, files in os.walk(source_dir):
         for file in files:
-            if file in EXCLUDED_FILES:
+            if should_exclude_file(file):
                 print(f"Skipping excluded file: {file}")
                 continue
             local_path = os.path.join(root, file)
@@ -218,7 +228,7 @@ def copy_s3_to_s3(source_bucket: str, source_prefix: str, dest_bucket: str, dest
                 continue
 
             filename = os.path.basename(key)
-            if filename in EXCLUDED_FILES:
+            if should_exclude_file(filename):
                 print(f"Skipping excluded file: {filename}")
                 continue
 
@@ -248,6 +258,9 @@ def prepare_checkpoints(sources: list[str], dest_path: str) -> None:
     """Prepare OlmOCR checkpoint(s) for deployment, with support for souping."""
     print(f"Preparing {'souped ' if len(sources) > 1 else ''}checkpoint from {len(sources)} source(s) to {dest_path}")
 
+    sources = [source.rstrip("/") for source in sources]
+    dest_path = dest_path.rstrip("/")
+
     # Detect architectures
     architectures = []
     for source in sources:
@@ -327,29 +340,55 @@ def prepare_checkpoints(sources: list[str], dest_path: str) -> None:
 
                 if file_path.endswith(".safetensors"):
                     sum_state = load_file(file_path, device="cpu")
+                    # Store original dtypes for each tensor
+                    original_dtypes = {k: v.dtype for k, v in sum_state.items()}
+                    # Upconvert to at least fp32 for accurate averaging
+                    for k in sum_state:
+                        if sum_state[k].dtype in (torch.float16, torch.bfloat16):
+                            sum_state[k] = sum_state[k].to(torch.float32)
+
                     for other_path in all_paths[1:]:
                         other_state = load_file(other_path, device="cpu")
                         if set(sum_state.keys()) != set(other_state.keys()):
                             raise ValueError(f"Key mismatch in {rel_path}")
                         for k in sum_state:
+                            # Upconvert other state to match sum_state dtype
+                            if other_state[k].dtype in (torch.float16, torch.bfloat16):
+                                other_state[k] = other_state[k].to(torch.float32)
                             sum_state[k] += other_state[k]
                         del other_state
+
                     n = len(all_paths)
                     for k in sum_state:
                         sum_state[k] /= n
+                        # Cast back to original dtype
+                        sum_state[k] = sum_state[k].to(original_dtypes[k])
                     save_file(sum_state, souped_path)
                 elif file_path.endswith(".bin"):
                     sum_state = torch.load(file_path, map_location="cpu")
+                    # Store original dtypes for each tensor
+                    original_dtypes = {k: v.dtype for k, v in sum_state.items()}
+                    # Upconvert to at least fp32 for accurate averaging
+                    for k in sum_state:
+                        if sum_state[k].dtype in (torch.float16, torch.bfloat16):
+                            sum_state[k] = sum_state[k].to(torch.float32)
+
                     for other_path in all_paths[1:]:
                         other_state = torch.load(other_path, map_location="cpu")
                         if set(sum_state.keys()) != set(other_state.keys()):
                             raise ValueError(f"Key mismatch in {rel_path}")
                         for k in sum_state:
+                            # Upconvert other state to match sum_state dtype
+                            if other_state[k].dtype in (torch.float16, torch.bfloat16):
+                                other_state[k] = other_state[k].to(torch.float32)
                             sum_state[k] += other_state[k]
                         del other_state
+
                     n = len(all_paths)
                     for k in sum_state:
                         sum_state[k] /= n
+                        # Cast back to original dtype
+                        sum_state[k] = sum_state[k].to(original_dtypes[k])
                     torch.save(sum_state, souped_path)
                 else:
                     print(f"Skipping unknown weight file: {rel_path}")
